import numpy as np
import torch
from functools import wraps

def inv_from_normalization(func):
    """
    Workflow of parameters in galmoss:
        Original domain 
            --> to_mapped --> 
        mapped domain 
            --> to_logged --> 
        logged domain
            --> to_normalized -->
        normalized domain and be load into optimizer
    The inv-workflow follows the same logic.

    Checks whether need to transform input values from normlization to 
    logged domain follows the attribute `value_range`. If value_range 
    is not None, which means the value in the func has already transformed 
    into (-1, 1) domain before, so the transition will be effective. 
    Otherwise, return input values. 

    Parameters
    ----------
    func: (parameter, *args, **kwargs) -> torch.tensor
        A function that receives values in (-1, 1)/non-normalized domain.

    Returns
    -------
        A function that can accept normalized or non-normalized values.
    """
    @wraps(func)
    def wrapper(
        cls,
        normalized_value: torch.tensor,
        *args,
        **kwargs
    ) -> torch.tensor:
        """
        Parameters
        ----------
        cls
            The class that owns the function.
            
        normalized_value
            The parameter value (maybe in (-1, 1) domain, depends on the
             `log` attribution).

        Returns
        -------
            The parameter value in mapped domain after conversion.
        """    
        value_range = cls.value_range #param_value_range
        
        if value_range != (None, None):
            return func(
                cls, 
                ((
                    torch.tanh(normalized_value) 
                        - cls.normalization_relationship[1]
                 ) 
                    * cls.normalization_relationship[0]
                ), 
                *args, 
                **kwargs
            ) 
        else:
          return func(cls, normalized_value, *args, **kwargs)
      
    return wrapper


def inv_from_log(func):
    """
    Workflow of parameters in galmoss:
        Original domain 
            --> to_mapped --> 
        mapped domain 
            --> to_logged --> 
        logged domain
            --> to_normalized -->
        normalized domain and be load into optimizer
    The inv-workflow follows the same logic.

    Checks whether need to transform input values from logged to 
    mapped domain follows the attribute `log`. If log is True, which 
    means the value in the func has already transformed into logarithmic 
    domain before, so the transition will be effective. Otherwise, return 
    input values.

    Parameters
    ----------
    func: (parameter, *args, **kwargs) -> torch.tensor
        A function that receives values in logarithmic/mapped domain.

    Returns
    -------
        A function that can accept logged or mapped values.
    """  
    @wraps(func)
    def wrapper(
        cls, 
        logged_value: torch.tensor, 
        *args,
        **kwargs
    ) -> torch.tensor:  
        """
        Parameters
        ----------
        cls
            The class that owns the function.
            
        logged_value
            The parameter value (maybe in logarithmic domain, depends on the
             `log` attribution).

        Returns
        -------
            The parameter value in normal domain after conversion.
        """             
        if cls.log:  
            return func(cls, torch.pow(10, logged_value), *args, **kwargs)  
        else:
            return func(cls, logged_value, *args, **kwargs)
           
    return wrapper


   
def inv_from_mapping_uncertainty(func):
    """
    Workflow of parameters in galmoss:
        Original domain 
            --> to_mapped --> 
        mapped domain 
            --> to_logged --> 
        logged domain
            --> to_normalized -->
        normalized domain and be load into optimizer
    The inv-workflow follows the same logic.

    Checks whether need to transform input values from mapped to 
    original domain follows the attribute `M0`, `pix_scale` and `angle`. 
    If M0 and pix_scale is not None or angle is True, which means the 
    value in the func has already transformed into mapped domain before, 
    so the transition will be effective. Otherwise, return input values. 

    Parameters
    ----------
    func: (parameter, *args, **kwargs) -> torch.tensor
        A function that receives uncertainty in mapped/non-mapped domain.

    Returns
    -------
        A function that can accept mapped or non-mapped uncertainty.
    """
    @wraps(func)
    def wrapper(cls, *args, **kwargs) -> torch.tensor:
        """
        Parameters
        ----------
        cls
            The class that owns the function.

        Returns
        -------
            The value attribute in original domain.
        """            
        mapped_uc = func(cls, *args, **kwargs)

        if cls.angle:
            inv_mapped_uc = mapped_uc * 180 / torch.pi
        elif cls.pix_scale is not None:
            inv_mapped_uc = mapped_uc * cls.pix_scale_device.squeeze()
        else:
            inv_mapped_uc = mapped_uc
            
        return inv_mapped_uc

    return wrapper
    
    
def inv_from_mapping(func):
    """
    Workflow of parameters in galmoss:
        Original domain 
            --> to_mapped --> 
        mapped domain 
            --> to_logged --> 
        logged domain
            --> to_normalized -->
        normalized domain and be load into optimizer
    The inv-workflow follows the same logic.

    Checks whether need to transform input values from mapped to 
    original domain follows the attribute `M0`, `pix_scale` and `angle`. 
    If M0 and pix_scale is not None or angle is True, which means the 
    value in the func has already transformed into mapped domain before, 
    so the transition will be effective. Otherwise, return input values. 

    Parameters
    ----------
    func: (parameter, *args, **kwargs) -> torch.tensor
        A function that receives values in mapped/non-mapped domain.

    Returns
    -------
        A function that can accept mapped or non-mapped values.
    """
    @wraps(func)
    def wrapper(cls, *args, **kwargs) -> torch.tensor:
        """
        Parameters
        ----------
        cls
            The class that owns the function.

        Returns
        -------
            The value attribute in original domain.
        """            
        mapped_value = func(cls, *args, **kwargs)
        if cls.M0 is not None:
            inv_mapped_value = mapped_value + cls.M0
        elif cls.angle:
            inv_mapped_value = mapped_value * 180 / torch.pi
            inv_mapped_value = transiform_eff_degrees(inv_mapped_value)
        elif cls.pix_scale is not None:
            inv_mapped_value = mapped_value * cls.pix_scale_device
        else:
            inv_mapped_value = mapped_value
        return inv_mapped_value

    return wrapper


def to_log(func): 
    """
    Workflow of parameters in galmoss:
        Original domain 
            --> to_mapped --> 
        mapped domain 
            --> to_logged --> 
        logged domain
            --> to_normalized -->
        normalized domain and be load into optimizer
    The inv-workflow follows the same logic.

    Checks whether need to transform input values to logged domain 
    follows the attribute `log`. If log is True, the transition will be 
    effective. Otherwise, return input values.

    Parameters
    ----------
    func: (parameter, *args, **kwargs) -> torch.tensor
        A function that returns value attributes, which possibly needs log
        transformation.

    Returns
    -------
        A function that returns a value attribute.
    """     
    @wraps(func)
    def wrapper(cls, *args, **kwargs) -> torch.tensor:
        """
        Parameters
        ----------
        cls
            The class that owns the function.

        Returns
        -------
            The value attribute in logarithmic domain (cls.log=True), 
            normal domain (cls.log=False), or None if value attributes 
            is None.
        """    
        Params = func(cls, *args, **kwargs)
        if Params is not None:
            if cls.log:  
                return torch.log10(Params)   
            else:
                return Params
        else:
            return None
        
    return wrapper
        
        
def to_normalization(func):
    """
    Workflow of parameters in galmoss:
        Original domain 
            --> to_mapped --> 
        mapped domain 
            --> to_logged --> 
        logged domain
            --> to_normalized -->
        normalized domain and be load into optimizer
    The inv-workflow follows the same logic.

    Checks whether need to normalize input values into (-1, 1) noramlized 
    domain follows the attribute `value_range`. If value_range is not None, 
    which means the value has constrains during the fitting pocess, so the 
    transition will be effective. Otherwise, return input values.    

    Parameters
    ----------
    func: (parameter, *args, **kwargs) -> torch.tensor
        A function that returns value attributes, which possibly needs 
        to be normalizationed.

    Returns
    -------
        A function that returns a value attribute.
    """
    @wraps(func)
    def wrapper(cls, *args, **kwargs) -> torch.tensor:
        """
        Parameters
        ----------
        cls: 
            The class that owns the function.

        Returns
        -------
            The value attribute in (-1, 1) domain (cls.value_range is 
            not None), normal domain (cls.value_range is None), or None 
            if value attributes is None.
        """
        
        Params = func(cls, *args, **kwargs)
        value_range = cls.value_range
        if Params is not None:
            if value_range != (None, None):  
                return (
                    torch.atanh(
                        (Params - value_range[1]) 
                        / (0.5 * (value_range[1] - value_range[0]))
                        + 1                       
                    )
                )
            else:
                return Params
        else:
            return None

    return wrapper


def to_mapping_log_with_value_range(func):
    """
    Workflow of parameters in galmoss:
        Original domain 
            --> to_mapped --> 
        mapped domain 
            --> to_logged --> 
        logged domain
            --> to_normalized -->
        normalized domain and be load into optimizer
    The inv-workflow follows the same logic.

    Checks whether need to transform input values from original to 
    mapped and then to logged domain. 

    Parameters
    ----------
    func: (parameter, *args, **kwargs) -> torch.tensor
        A function that receives values and value ranges in original 
        domain.

    Returns
    -------
        A function that can accept original values and value ranges.
    """
    @wraps(func)
    def wrapper(
        cls, 
        parameters,
        value_range,
        M0,
        pix_scale,
        angle,
        *args, 
        **kwargs
    ) -> torch.tensor:
        """

        Parameters
        ----------
        cls
            The class that owns the function.
            
        parameters
            The parameter value in original domain.

        value_range
            The parameter value in original domain.

        M0
            The magnitude zero point.

        pix_scale
            The arcsec to pixel ratio.

        angle
            To show whether the parameter is an angle or not.

        Returns
        -------
            The parameter value in logged domain after conversion.
        """   
        if value_range != (None, None):
            if M0 is not None:
                mapped_value_range = [value_range[0] - M0, 
                                      value_range[1] - M0]
            elif angle:
                mapped_value_range = [value_range[0] * torch.pi / 180, 
                                      value_range[1] * torch.pi / 180]
            elif pix_scale is not None:
                mapped_value_range = [value_range[0] / pix_scale, 
                                      value_range[1] / pix_scale]
            else:
                mapped_value_range = value_range
        else:
            mapped_value_range = value_range

        if M0 is not None:
            mapped_parameters = parameters - M0
        elif angle:
            mapped_parameters = (transiform_eff_degrees(parameters) 
                                 * torch.pi / 180)
        elif pix_scale is not None:
            if isinstance(pix_scale, (int, float)):
                mapped_parameters = parameters / pix_scale
            else:
                print(pix_scale)
                mapped_parameters = parameters / pix_scale.reshape(-1, 1)
        else:
            mapped_parameters = parameters                      

        if cls.log:
            log_parameters = np.log10(mapped_parameters)
            if mapped_value_range == (None, None):
                log_value_range = mapped_value_range
            else:
                log_value_range = (np.log10(mapped_value_range[0]), 
                                   np.log10(mapped_value_range[1]))
        else:
            log_parameters = mapped_parameters
            log_value_range = mapped_value_range

        return func(cls, log_parameters, log_value_range, *args, **kwargs)          

    return wrapper


def redim(func):
    """
    To transit input values into torch.tensor, and resize to (param_length, 
    param_dim). param_length equals the number of total galaxies. 

    Parameters
    ----------
    func: (parameter, *args, **kwargs) -> torch.tensor
        A function that receives values in whatever data types.

    Returns
    -------
        A function that can accept whatever data types.
    """
    @wraps(func)
    def wrapper(
        cls, 
        parameters,
        *args, 
        **kwargs
    ) -> torch.tensor:
        """
        Parameters
        ----------
        cls
            The class that owns the function.
            
        parameters
            The parameter value that being input in the object by users.

        Returns
        -------
            The parameter value in torch.tensor, and being resize into 
            (param_length, param_dim).
        """        
        if not isinstance(parameters, torch.Tensor):
            parameters = torch.from_numpy(np.asarray(parameters)).squeeze() 
        else:
            parameters = parameters.squeeze()
        
        if cls.dim == "1d":
            if parameters.ndim not in [0, 1]:
                raise ValueError("Invalid input dimensions for SingleDim mode.")
            parameters = parameters.view(-1, 1) 
        elif cls.dim == "2d":
            if parameters.ndim != 2:
                raise ValueError("Invalid input dimensions for MultiDim mode,"
                                 "expected 2D tensor.")
        else:
            raise ValueError("Invalid mode specified.")

        cls.total_initial_value_cpu_tensor = parameters
        return func(cls, parameters, *args, **kwargs) 
    return wrapper


def transiform_eff_degrees(degrees):
    """
    To map the degrees out of the value range into the effective value 
    inside the value range (-180~180).

    Parameters
    ----------
    degrees
        The degrees, that may have values out of the value range.

    Returns
    -------
        The degrees all inside the value range.    
    """
    degrees = degrees % 360
    degrees = torch.where(degrees > 180, degrees - 360, degrees)
    degrees = torch.where(degrees > 90, degrees - 180, degrees)
    degrees = torch.where(degrees < -90, degrees + 180, degrees)
    return degrees




from .Parameter.basic import Parameters as p
from .data import Profiles, DataSet
from .profile.basic import LightProfile
from .fitting import Fitting
from .profile import light_profile as lp
__version__ = '2.0.7'


import numpy as np
import torch
import pandas as pd
from typing import Union, Tuple

from galmoss.Parameter import decorater  


class ParamsRepo:
     
    def __init__(self):   
        """
        This repository stores various types of values with specific
        naming conventions:

        1. 'total' vs 'batched': 
            'batched' variables change with each galaxy batch during
            fitting, whereas 'total' variables accumulate values across
            batches, completing only after the entire fitting process.

        2. Domain types - 'mapped', 'normalized', or unspecified: 
            Variables are initially in the 'original input domain'. 
            'mapped' variables are in the 'profile equation domain' 
            relevant for PA, scale-related variables (e.g., eff_r), and 
            mag. 'normalized' variables, often used in optimizers, are 
            scaled to (-1~1) when range constraints apply.

        3. Value states - 'initial', 'updating', 'best': 
            'initial' variables are the starting value in the parameter
            object; 'updating' variables change during fitting; 'best' 
            variables are the final fitted value.

        4. Storage formats - 'cpu', 'numpy', or unspecified: 
            Variables with 'cpu' are tensors on CPU; with 'numpy' are 
            numpy.array on CPU; unspecified ones are tensors on 
            user-defined devices, typically GPU (cuda) by default.

        Value names and comments indicate their nature, e.g., 
        'self.total_mapped_initial_val_cpu' signifies initial total values 
        in the original input domain, stored as tensors on CPU. A value 
        dictionary "self.value_dict" allows quick indexing. Changes here 
        require corresponding updates in the `value` function's ValueError 
        information.
        """        
        if self.log:
            self.total_mapped_inital_val_cpu = torch.pow(10, self._parameters)
        else:
            self.total_mapped_inital_val_cpu = self._parameters
        self.value_dict = {
            "updating_model": "batched_broadcasted_updating_val", 
            "best_model": "batched_broadcasted_mapped_best_val",
            "load_best_in_optim_list": "total_normalized_best_val_cpu",
            "load_best_in_jacobian": "total_mapped_best_val_cpu",
            "load_initial_in_jacobian": "total_mapped_ini_value", 
            "load_initial_in_optim_list": "total_normalized_ini_value", 
            "initial_model": "total_broadcasted_mapped_inital_val",
        }    
            
    def value(self, mode: str = "updating_model") -> torch.tensor:
        """
        Return the value guided by the 'mode' argument.
        
        Parameters
        ----------
        mode
            The string representing which value to choose from self.value_dict.
        """              
        if mode not in self.value_dict:
            raise ValueError(
                "Wrong mode text, should be one of: updating_model, best_model, "
                "load_best_in_optim_list, load_best_in_jacobian, "
                "load_initial_in_optim_list, or initial_model"
            )
        
        return getattr(self, self.value_dict[mode])
 
    @property
    def batched_broadcasted_updating_val(self):
        """
        Return the updating value in each sub-fitting batch, mapped to 
        the profile equation domain 
        (e.g., mag' = mag - M0, eff_r' = eff_r / pix_scale), and 
        broadcasted for calculations. This is typically used to generate
        the model image during the fitting process.

        This attribute can also be accessed via 
        param.value("updating_model").
        """   
        if self.batched_updating_value is not None:
            if self.dim == "1d":
                return self.batched_updating_value.unsqueeze(1) 
        else:
            return None        

    @property
    def total_broadcasted_mapped_inital_val(self, device: str = "cpu"): 
        """
        Return the total initial value, mapped to the profile equation
        domain (e.g., mag' = mag - M0, eff_r' = eff_r / pix_scale), and
        broadcasted for calculations. The data is loaded on the CPU by
        default, but can be changed to other devices using the 'device' 
        argument. This is typically used to generate the initial model 
        image for a quick review.
        
        This attribute can also be accessed by 
        param.value("initial_model").
        """         
        if self.total_mapped_inital_val_cpu is not None:
            return self.total_mapped_inital_val_cpu.unsqueeze(1).to(device)
        else:
            return None        

    @property
    def total_initial_value_numpy(self): 
        """
        Return the total initial value in the original input domain,
        with shape (param_length, dim_length). This is typically used to
        save the fitted parameter values for fixed parameters, as they
        only have an initial value.
        """          
        if self.total_initial_value_cpu_tensor is not None:
            return self.total_initial_value_cpu_tensor.numpy()
        else:
            return None        
                 
    @property
    @decorater.to_normalization
    def total_normalized_ini_value(self):
        """
        Return the total initial value in the normalized domain. This is 
        typically for loading into the optim_list that will later be
        loaded into optimizer in the fitting process. 

        This attribute can also be accessed by 
        param.value("load_initial_in_optim_list").
        """              
        return self._parameters

  
    @property
    @decorater.inv_from_log
    def total_mapped_ini_value(self):
        """
        Return the total initial value in the normalized domain. This is 
        typically for loading into the optim_list that will later be
        loaded into optimizer in the fitting process. 

        This attribute can also be accessed by 
        param.value("load_initial_in_optim_list").
        """              
        return self._parameters
    
      
    @property
    def total_mapped_best_val_cpu(self): 
        """
        Return the total best value, mapped to the profile equation
        domain (e.g., mag' = mag - M0, eff_r' = eff_r / pix_scale), and 
        broadcasted for calculations. This is typically used to generate
        the model image during the fitting uncertainty estimation 
        process using co-variance matrix.
        
        This attribute can also be accessed by 
        param.value("load_best_in_jacobian").
        """       
        if hasattr(self, "total_mapped_best_value"):
            if self.dim == "1d":
                return self.total_mapped_best_value.detach().cpu()
        else:
            return None

    @property 
    @decorater.inv_from_mapping
    def total_best_value(self):
        """
        Return the total best value in the original input domain,
        with shape (param_length, dim_length), in numpy.array. 
        """       
        if hasattr(self, "total_mapped_best_value"):
            return self.total_mapped_best_value.detach()
        else:
            return None

    @property 
    @decorater.inv_from_mapping
    def total_bsp_uncertainty(self):
        """
        Return the total uncertainty in the original input domain,
        with shape (param_length, dim_length), in torch.Tensor. 
        """       
        if hasattr(self, "batched_mapped_best_value"):
            return self.batched_mapped_best_value.detach()
        else:
            return None
        
    @property 
    def total_best_value_numpy(self):
        """
        Return the total best value in the original input domain,
        with shape (param_length, dim_length), in numpy.array. This is 
        typically used to save the fitted values for variable parameters.
        """       
        if hasattr(self, "total_best_value"):
            return self.total_best_value.cpu().numpy() 
        else:
            return None
        
    @property
    def batched_broadcasted_mapped_best_val(self):
        """
        Return the best value in each sub-fitting batch, mapped to 
        the profile equation domain 
        (e.g., mag' = mag - M0, eff_r' = eff_r / pix_scale), and 
        broadcasted for calculations. This is typically used to generate
        the model image in the result image saving process.

        This attribute can also be accessed via 
        param.value("best_model").
        """           
        if self.batched_mapped_best_value is not None:
            if self.dim == "1d":
                return self.batched_mapped_best_value.unsqueeze(1)
        else:
            return None

    @property
    @decorater.to_normalization
    @decorater.to_log
    def total_normalized_best_val_cpu(self):
        """
        Return the total best value in the normalized domain. This is 
        typically for loading into the optim_list that will later be
        loaded into optimizer in the bootstrapping process. 

        This attribute can also be accessed by 
        param.value("load_best_in_optim_list").
        """     
        if hasattr(self, "total_mapped_best_value"):
            return self.total_mapped_best_value.detach().cpu()
        else:
            return None
                     
    @property
    def total_uncertainty_numpy(self):
        """
        Return the total uncertainty value in the original input domain,
        with shape (param_length, dim_length). This is typically used to
        save the fitting uncertainty for variable parameters.
        """  
        if hasattr(self, "uncertainty"):
            return self.uncertainty.detach().cpu().numpy() 
        else:
            return None

  
class Parameters(ParamsRepo):

    def __init__(
        self, 
        parameters: Union[np.ndarray, torch.Tensor, float, pd.Series], 
        value_range:  Union[Tuple[float, float], None] = (None, None),
        step_length: float = 0.015,
        M0: float = None,
        pix_scale: Union[float, None] = None,
        angle: bool = False,
        fit: bool = True,
        log: bool = False,  
        dim: str = "1d"
    ):  
        """
        In GALMOSS, each parameter is represented as an individual object 
        instantiated from the Parameters class. The Parameters class is a 
        fundamental class in GALMOSS. Other classes such as Dataset, 
        Profiles, and fitting classes, can access the attributes of the 
        Parameters class.

        Parameters
        ----------
        parameters : Union[np.ndarray, torch.Tensor, float, pd.Series]
            The original input value of the parameter.
        value_range : Union[Tuple[float, float], None], optional
            Constrains the fitting range.
        step_length : float, optional
            The optimization step length in gradient descent (learning rate 
            in PyTorch's optimizer), by default 0.015.
        M0 : float, optional
            The magnitude zero point for magnitude parameters.
        pix_scale
            The conversion factor between arcseconds and pixels, applicable 
            only to scale-related parameters like eff_r.
        angle
            Indicates if the parameter is angle-related.
        fit
            Indicates if the parameter is variable or fixed during fitting.
        log
            Determines if the parameter should be fitted in log space.
        dim 
            Specifies if the parameter is two-dimensional ('1d' or '2d'). 
            This argument will only be changed in user-defined profile that 
            is very special! 
            
        Attributes
        ----------
        normalization_relationship
            Calculates the normalization relationship based on the provided 
            value_range.

        dim2_length
            In built-in profiles, this attribute is 1. For 2-D parameters, 
            this attribute changes to 2.

        param_length
            The length of the parameter, which is equal to the number of 
            galaxies.            
        """                
        if sum([M0 is not None, 
                pix_scale is not None, angle is not False]) > 1:
            raise ValueError("A parameter can only be one of Magnitude, PA or "
                             "scale_related parameter.")
        if dim in ["1d", "2d"]:
            self.dim = dim
        else:
            raise ValueError(f"The dim '{dim}' is not valid.")

        self.M0 = M0
        if pix_scale is not None:
            self.pix_scale = np.array(pix_scale)
        else:
            self.pix_scale = pix_scale
        self.angle = angle
        self.fit = fit
        self.log = log
        self.step_length = step_length
        self._parameters, self.value_range = self.mapping(parameters, 
                                                          value_range,
                                                          M0, self.pix_scale, 
                                                          angle)
        self.dim2_length = self._parameters.shape[1]
        self.param_length = self._parameters.shape[0]

        if value_range != (None, None):
            self.normalization_relationship = (
                (self.value_range[1] - self.value_range[0]) / 2,
                1 - self.value_range[1] / ((self.value_range[1] - self.value_range[0]) / 2)
            )
            condition = (
                torch.max(self._parameters) <= self.value_range[1] and
                torch.min(self._parameters) >= self.value_range[0]
            )
            assert condition, ("Data is out of the specified range. Please check"
                                "the input values.")
        else:
            self.normalization_relationship = (1, 0)

        super().__init__()
   
    
    @decorater.redim
    @decorater.to_mapping_log_with_value_range
    def mapping(self, 
                parameter: torch.tensor, 
                value_range: Tuple[float, float], 
                *args, 
                **kwargs):
        """
        Returns the mapped parameter and value range. If log=True, this 
        function returns the log of the mapped values.

        Parameters
        ----------
        parameter
            The parameter value in the original input domain.
        value_range
            Constrains the fitting range, with values in the original 
            input domain.
        *args & kwargs
            Additional arguments, can only be M0, pix_scale, or angle. 
            M0 is the magnitude zero point for magnitude parameters. 
            pix_scale is the conversion factor between arcseconds and 
            pixels, applicable only to scale-related parameters like 
            eff_r. angle indicates if the parameter is angle-related.
        """
        return parameter, value_range

    @decorater.inv_from_normalization 
    @decorater.inv_from_log          
    def update_updating_val(self, value: torch.tensor):
        """
        Returns the mapped parameter and value range. If log=True, this
        function returns the log of the mapped values.

        Parameters
        ----------
        parameter
            The parameter value in the original input domain.
        value_range
            Constrains the fitting range, with values in the original
            input domain.
        *args&**kwargs
            Additional arguments, can only be M0, pix_scale, or angle.
            M0 is the magnitude zero point for magnitude parameters. 
            pix_scale is the conversion factor between arcseconds and 
            pixels, applicable only to scale-related parameters like 
            eff_r. angle indicates if the parameter is angle-related.
        """           
        self.batched_updating_value = value

    def update_best_val(self, index: Union[torch.tensor, bool] = None):
        """
        Update the best value during the optimization process. Not every 
        iteration in the optimization process results in a better outcome
        with a lower chi-square. Therefore, updates are made only in 
        iterations that successfully reduce the chi-square, guided by 
        the given index.

        Parameters
        ----------
        index
            Specifies the position in the best value matrix to be
            refreshed. If None, refreshes the entire matrix.
        """               
        if index is None:
            self.batched_mapped_best_value = self.batched_updating_value
        else:
            self.batched_mapped_best_value[index] = self.batched_updating_value[index]
   
    def best_value_append(self, batched_mapped_best_value: torch.tensor):
        """
        Append the best fitted values for each batch of galaxies after
        the completion of each batch fitting process.

        Parameters
        ----------
        batched_mapped_best_value
            The best fitted value for each batch of galaxies.
        """         
        if hasattr(self, 'total_mapped_best_value'):
            self.total_mapped_best_value = torch.concat(
                (self.total_mapped_best_value, batched_mapped_best_value), dim=0)                  
        else:
            self.total_mapped_best_value = batched_mapped_best_value


    @decorater.inv_from_mapping_uncertainty
    def inv_from_mapping_uc(self, data):
        """
        Return the total uncertainty in the original input domain,
        with shape (param_length, dim_length), in numpy.array. 
        """       
        return data
 
 
    # @decorater.inv_from_normalization_and_log_uncertainty
    # def inv_from_normalization_and_log_uc(self, uc, data):
    #     """
    #     Return the total uncertainty in the original input domain,
    #     with shape (param_length, dim_length), in numpy.array. 
    #     """       
    #     return uc     
    
      
    def calculate_uncertainty(self, proj_uc):
        """
        This function is for calculating the final uncertainty of parameters. 
        Inside function "cm_uncertainty", we get the projected uncertianty 
        `self.uncertainty`of parameters, so we need the error propagation 
        equation to expand.
        
        Parameters
        ----------
        proj_uc
            The projected uncertianty of parameters           
        """    
        if self.log:
            real_uc = (
                torch.exp(
                    (proj_uc - self.normalization_relationship[1]) 
                    * self.normalization_relationship[0]
                ) 
                * self.normalization_relationship[0] 
                * proj_uc
            )
        else:
            real_uc = self.normalization_relationship[0] * proj_uc
            
        if hasattr(self, 'uncertainty_gpu'):
            self.uncertainty = torch.concat(
                (self.uncertainty_gpu, real_uc)
                , dim=0
                )                   
        else:
            self.uncertainty = real_uc
            
    def __repr__(self) -> str:
        return self.__class__.__name__    


from typing import Tuple, Union
import numpy as np
import torch
from functools import wraps

from galmoss.Parameter.basic import Parameters


def revise_grid(func):
    """
    Transform the grid to galaxy-centered, and then add a shift to avoid 
    the values near to zero. Grid values which are near to zero may cause 
    NaN or Inf in calculations.

    Parameters
    ----------
    func : (profile, *args, **kwargs) -> torch.tensor
        A function that receives a original grid.

    Returns
    -------
        A function that can accept grids.
    """    
    @wraps(func)
    def wrapper(
        cls,
        grid,
        mode,
        *args,
        **kwargs
    ) -> np.ndarray:
        """
        Parameters
        ----------
        cls
            The class that owns the function.
        grid
            The coordinates in the original domain.
        mode:
            The value mode, which are used to call 
            various mode of parameter values.

        Returns
        -------
            The grid being transformed and revise from zero.
        """
        y = grid[0] - (grid[0].shape[0] - cls.cen_y.value(mode)) + 1e-6
        x = grid[1] - cls.cen_x.value(mode) + 1e-6

        grid = [x, y]         
        return func(cls, grid, mode, *args, **kwargs)

    return wrapper


class LightProfile:
    
    def __init__(self): 
        """
        Abstract base class for an light-profile.

        Each light profile has an equation, which represents the distribution 
        of surface brightness. The model image can be calculated given an 
        2D grid.
        
        Attribute
        ---------
        psf
            To control whether this profile will be convolved with PSF 
            imagec later.
        saving
            To control whether the model image of this profile will be
            saved in image block in the end of fitting process.
        """           
        self.psf = True
        self.saving = True
        
    def image_via_grid_from(
        self, 
        grid: Union[torch.tensor, torch.tensor],
        mode: str):
            """
            Returns the light profile's image calculated from a 2D grid.

            Parameters
            ----------
            grid
                The galaxy-centered grid.
            mode
                Determine which kind of values will be called in 
                parameter object.

            Returns
            -------
            image
                The model image calculated with the grid.
            """
            raise NotImplementedError()    

    def manage_params(self, name: str) -> dict:
        """
        To manage the parameters in this profile. The parameters will be 
        divide into fixed/variable params based on the value of their 
        attribute `fit`, and be stored in correspoding dictionaries. These
        dictionaries can be called by `variable_dic` and `fixed_dic`. This 
        function will also check the length of these parameters, which 
        will raise error if they are not concise.

        Parameters
        ----------
        name
            The user-defined name of this profile, which will be used in name 
            the keys of each values in dictionaries like "profile_name" + "_"
            + "param_name" 

        """        
        variable_dic = {}
        fixed_dic = {}
        param_length = []
        for param_name in dir(self):
            param = getattr(self, param_name)            
            if isinstance(param, Parameters):
                param.param_name = param_name
                param_length.append(param.param_length)
                idx = name + "_" + param_name
                if param.fit:
                    variable_dic[idx] = param 
                else:
                    fixed_dic[idx] = param
        assert all(x == param_length[0] 
                   for x in param_length), ("the length of parameters"
                                            "are not the same!")
        self.variable_dic = variable_dic
        self.fixed_dic = fixed_dic


class Ellip(LightProfile):
     
    def __init__(
        self, 
        cen_x: Parameters, 
        cen_y: Parameters, 
        pa: Parameters,
        axis_r: Parameters,
        box: Parameters
    ):
        """
        Abstract base class for an radial profile.

        Each radial profile has a transition from grid to radial. 
        
        Parameters
        ---------
        cen_x
            The galaxy center in x-xis in arc-second.
        cen_y
            The galaxy center in y-xis in arc-second.
        pa
            The position angle of the ellipse.
        axis_r
            The axis ratio of the ellipse.
        box
            The boxness of the ellipse.    
        """        
        super().__init__()
        self.cen_x = cen_x
        self.cen_y = cen_y
        self.pa = pa
        self.axis_r = axis_r    
        
        if box == None:
            self.box = Parameters(parameters=torch.zeros(cen_x.param_length), 
                                  fit=False)   
        else:
            self.box = box  
        
    def R_box(self, mode):
        """
        The geometric correction factor in Ie to account for deviations 
        from a perfect ellipse, influenced by the level of diskiness or 
        boxiness. Specifically, when box = 0, indicating a perfect 
        ellipse, R_box = 1, implying no geometric correction. 
        """         
        beta = (torch.exp(torch.lgamma(1/(2 + self.box.value(mode)))) 
                * torch.exp(torch.lgamma(1/(2 + self.box.value(mode)))) 
                / torch.exp(torch.lgamma(2/(2 + self.box.value(mode)))))
        
        return torch.pi * (2 + self.box.value(mode)) / (2 * beta)

    @revise_grid
    def make_radius(self, 
                    grid: Tuple[torch.Tensor, torch.Tensor], 
                    mode="updating_model"):
        """
        Calculate the radius from the grid. 
        
        rmaj = |cos(θ)(x - xc)| + |sin(θ)(y - yc)|
        rmin = |- sin(θ)(x - xc)| + |cos(θ)(y - yc)|
        
        r = (rmaj**(B+2) + (rmin/q)**(B+2))**(1/(B+2))

        Parameters
        ---------
        grid
            The grid, transit to galaxy-centered and revise the value close 
            to zero by the decorator.
        mode
            Determine which kind of values will be called in 
            parameter object.  
        """   
              
        y_grid = grid[0]
        x_grid = grid[1]  
        self.x_maj = torch.abs(x_grid * torch.cos(self.pa.value(mode))
                                + y_grid * torch.sin(self.pa.value(mode)))
        self.x_min = torch.abs(- x_grid * torch.sin(self.pa.value(mode)) 
                               + y_grid * torch.cos(self.pa.value(mode)))
        self.radius = torch.pow(torch.pow(self.x_maj, 
                                          self.box.value(mode) + 2) 
                                + torch.pow((self.x_min / self.axis_r.value(mode)), 
                                            (self.box.value(mode) + 2)), 
                                1 / (self.box.value(mode) + 2))

        return self.radius
 
    def image_via_grid_from(
        self, 
        grid: Union[torch.tensor, torch.tensor], 
        mode: str = "updating_model"):
        """
        Returns the light profile's image calculated from a 2D grid. In 
        radial profiles, the profiles are equations with radial, so grid
        need to transit to radial to do the further calculations.

        Parameters
        ----------
        grid
            The galaxy-centered grid.
        mode
            Determine which kind of values will be called in 
            parameter object.

        Returns
        -------
        image
            The model image calculated with the grid.
        """
        return self.image_via_radii_from(self.make_radius(grid, mode), mode)     
     
    def image_via_radii_from(
        self, 
        radius: torch.tensor, 
        mode: str):
        """
        Returns the light profile's image calculated from a radius matrix. 

        Parameters
        ----------
        radius
            The radius of the equivalent circle.
        mode
            Determine which kind of values will be called in 
            parameter object.

        Returns
        -------
        image
            The model image calculated with the radius.
        """
        raise NotImplementedError()




from galmoss.Parameter.basic import Parameters
import torch
from typing import Union
import numpy as np
from galmoss.profile.basic import LightProfile, Ellip


class Sersic(Ellip, LightProfile):
    def __init__(
    self, 
    cen_x: Parameters, 
    cen_y: Parameters, 
    pa: Parameters,
    axis_r: Parameters,
    
    eff_r: Parameters,
    mag: Parameters,
    ser_n: Parameters,
    
    box: Parameters = None,
    ):
        """
        The Sersic profile. 
        
        Parameters
        ----------
        cen_x
            The galaxy center in x-xis in arc-second.
        cen_y
            The galaxy center in y-xis in arc-second.
        pa
            The position angle of the ellipse.
        axis_r
            The axis ratio of the ellipse.      
        eff_r
            The effective radius, which encompasses half of the total profile flux.
        mag
            The total magnitude of the galaxy.
        ser_n
            The sersic index, which dictates the profile's curvature
        box
            The boxness of the ellipse.  
        """  
        super().__init__(cen_x=cen_x, cen_y=cen_y, pa=pa, axis_r=axis_r, box=box)
        
        self.eff_r = eff_r
        self.mag = mag
        self.ser_n = ser_n 

    def sersic_constant(self, ser_n: torch.tensor) -> float:
        """
        A parameter derived from Sersic index which ensures that effective radius contains 50% of the profile's total integrated light.
        """
        greater_condition = ser_n > 0.36
        greater_values = (
            (2 * ser_n)
            - (1.0 / 3.0)
            + (4.0 / (405.0 * ser_n))
            + (46.0 / (25515.0 * ser_n**2))
            + (131.0 / (1148175.0 * ser_n**3))
            - (2194697.0 / (30690717750.0 * ser_n**4))
        )

        lesser_values = (
            0.01945 
            - 0.8902 * ser_n
            + 10.95 * ser_n**2
            - 19.67 * ser_n**3
            + 13.43 * ser_n**4
        )

        return torch.where(greater_condition, greater_values, lesser_values)

    
    def image_via_radii_from(self, 
                             radius:torch.tensor, 
                             mode: str = "updating_model"):
        """
        Returns the model image of the sersic profile from radius.

        Parameters
        ----------
        radius
            The radius of the equivalent circle.
        mode
            Determine which kind of values will be called in parameter 
            object.
        """              
        I_e = (
            (torch.pow(10, (-0.4 * self.mag.value(mode)))*self.R_box(mode))
             / (torch.pow(self.eff_r.value(mode), 2)*2*torch.pi
             * self.ser_n.value(mode)
             * torch.exp(self.sersic_constant(self.ser_n.value(mode)))
             * self.axis_r.value(mode)
             * (torch.exp(torch.lgamma(2 * self.ser_n.value(mode))))  
             / torch.pow(10, 
                         2 * self.ser_n.value(mode) 
                         *torch.log10(self.sersic_constant(self.ser_n.value(mode)))
                         )
             )
             )
        # np.save("/data/public/ChenM/MIFIT/MANGA/data/g/final_paper_data/ie.npy", I_e.detach().cpu().numpy())
        # exit()
        return (I_e 
                * torch.exp(-self.sersic_constant(self.ser_n.value(mode))
                            * ( torch.pow(radius / self.eff_r.value(mode), 
                                          1.0 / self.ser_n.value(mode))
                               - 1)
                            )
                )

    
class ExpDisk(Ellip, LightProfile):
    def __init__(
    self, 
    cen_x: Parameters, 
    cen_y: Parameters, 
    pa: Parameters,
    axis_r: Parameters,
    
    mag: Parameters,
    scale_l: Parameters,
    
    box: Parameters=None,
    ):
        """
        The Exponential disk profile. 

        Parameters
        ----------
        cen_x
            The galaxy center in x-xis in arc-second.
        cen_y
            The galaxy center in y-xis in arc-second.
        pa
            The position angle of the ellipse.
        axis_r
            The axis ratio of the ellipse.      
        mag
            The total magnitude of the galaxy.
        scale_l
            The scale-length of the disk.        
        box
            The boxness of the ellipse.  
        """  
        super().__init__(cen_x=cen_x, cen_y=cen_y, pa=pa, axis_r=axis_r, box=box)
        self.mag = mag
        self.scale_l = scale_l
        
        
    def image_via_radii_from(self, radius, mode="updating_model"):
        """
        Returns the model image of the Exponential disk profile from radius.

        Parameters
        ----------
        radius
            The radius of the equivalent circle.
        mode
            Determine which kind of values will be called in parameter 
            object.
        """              

        I_0 = (
            (torch.pow(10, -0.4 * self.mag.value(mode))
             * self.R_box(mode)*(self.con.value(mode)-1))
            /(2 * torch.pi * self.axis_r.value(mode) 
              * torch.pow(self.scale_l.value(mode), 2)))
        
        return I_0 * torch.exp(-radius/self.scale_l.value(mode))
                        

class Ferrer(Ellip, LightProfile):
    def __init__(
    self, 
    cen_x: Parameters, 
    cen_y: Parameters, 
    pa: Parameters,
    axis_r: Parameters,
    
    mag: Parameters,
    trunc_r: Parameters,
    trunc_a: Parameters,
    trunc_b: Parameters,

    box: Parameters=None,
    ):
        """
        The Ferrer profile. 

        Parameters
        ----------
        cen_x
            The galaxy center in x-xis in arc-second.
        cen_y
            The galaxy center in y-xis in arc-second.
        pa
            The position angle of the ellipse.
        axis_r
            The axis ratio of the ellipse.      
        mag
            The total magnitude of the galaxy.
        trunc_r
            The the outer truncation radius.     
        trunc_a
            The parameters governing the slopes of the core.
        trunc_b
            The parameters governing the slopes of the truncation.                         
        box
            The boxness of the ellipse.  
        """          
        super().__init__(cen_x=cen_x, cen_y=cen_y, pa=pa, axis_r=axis_r, box=box)
        self.mag = mag
        self.trunc_r = trunc_r
        self.trunc_a = trunc_a
        self.trunc_b = trunc_b

    def image_via_radii_from(self, radius, mode="updating_model"):
        """
        Returns the model image of the Ferrer profile from radius.

        Parameters
        ----------
        radius
            The radius of the equivalent circle.
        mode
            Determine which kind of values will be called in parameter 
            object.
        """              
        radius[radius>self.trunc_r.value(mode)] = self.trunc_r.value(mode)
        
        I_0 = (
            (
                torch.pow(10, (-0.4 * self.mag.value(mode)))
                * self.R_box(mode)
            )
               / (
                   torch.exp(torch.lgamma(self.trunc_a.value(mode))) 
                 * torch.exp(torch.lgamma(1 + 2/(2 - self.trunc_b.value(mode)))) 
                 / torch.exp(torch.lgamma(self.trunc_a.value(mode)
                                          + 1 
                                          + 2 / (2 - self.trunc_b.value(mode))
                                          )
                             ) 
                 * self.trunc_a.value(mode) 
                 * torch.pow(self.trunc_r.value(mode), 2) 
                 * self.axis_r.value(mode)
                 *torch.pi
                 )
               )
        return I_0 * torch.pow(1 - torch.pow(radius/self.trunc_r.value(mode), 
                                             2 - self.trunc_b.value(mode)
                                             ),
                               self.trunc_a.value(mode)
                               )
  
    
class King(Ellip, LightProfile):
    def __init__(
    self, 
    cen_x: Parameters, 
    cen_y: Parameters, 
    pa: Parameters,
    axis_r: Parameters,
    
    inten: Parameters,
    core_r: Parameters,
    trunc_r: Parameters,
    pow_n: Parameters,
    
    box: Parameters=None,
    ):
        """
        The King profile. 

        Parameters
        ----------
        cen_x
            The galaxy center in x-xis in arc-second.
        cen_y
            The galaxy center in y-xis in arc-second.
        pa
            The position angle of the ellipse.
        axis_r
            The axis ratio of the ellipse.      
        inten
            The central surface brightness parameter.
        core_r
            Core radius, signifies the scale at which the density starts 
            to deviate from uniformity.     
        trunc_r
            Truncation radius, marks the boundary of the cluster.
        pow_n
            The global power-law factor, dictates the rate at which the 
            density declines with distance from the center.             
        box
            The boxness of the ellipse.  
        """               
        super().__init__(cen_x=cen_x, cen_y=cen_y, pa=pa, axis_r=axis_r, box=box)
        self.inten = inten
        self.core_r = core_r
        self.trunc_r = trunc_r
        self.pow_n = pow_n


    def image_via_radii_from(self, radius, mode="updating_model"):
        """
        Returns the model image of the King profile from radius.

        Parameters
        ----------
        radius
            The radius of the equivalent circle.
        mode
            Determine which kind of values will be called in parameter 
            object.
        """              
        radius[radius>self.trunc_r.value(mode)] = self.trunc_r.value(mode)
        partA = torch.pow(
            (1 + torch.pow(
                 (self.trunc_r.value(mode)/self.core_r.value(mode)), 
                 2
                 )
             ), 
            1 /self.pow_n.value(mode)
            )
        partB = torch.pow(
            (1 + torch.pow(
                (radius/self.core_r.value(mode)), 
                2
                )
             ), 
            1 / self.pow_n.value(mode)
            )

        return (self.inten.value(mode) 
                * torch.pow(1 - (1/partA), -self.pow_n.value(mode)) 
                * torch.pow(1/partB - 1/partA, self.pow_n.value(mode)))


class Gaussian(Ellip, LightProfile):
    def __init__(
    self, 
    cen_x: Parameters, 
    cen_y: Parameters, 
    pa: Parameters,
    axis_r: Parameters,
    
    inten: Parameters,
    fwhm: Parameters,
    
    box: Parameters=None,
    ):
        """
        The Gaussian profile. 

        Parameters
        ----------
        cen_x
            The galaxy center in x-xis in arc-second.
        cen_y
            The galaxy center in y-xis in arc-second.
        pa
            The position angle of the ellipse.
        axis_r
            The axis ratio of the ellipse.      
        inten
            The central surface brightness parameter.
        fwhm
            The Full Width at Half Maximum.                 
        box
            The boxness of the ellipse.  
        """           
        super().__init__(cen_x=cen_x, cen_y=cen_y, pa=pa, axis_r=axis_r, box=box)
        self.fwhm = fwhm
        self.inten = inten
    

    def image_via_radii_from(self, radius, mode="updating_model"):
        """
        Returns the model image of the Gaussion profile from radius.

        Parameters
        ----------
        radius
            The radius of the equivalent circle.
        mode
            Determine which kind of values will be called in parameter 
            object.
        """                  
        return (self.inten.value(mode) 
                * torch.exp(
                    -torch.pow(radius, 2)
                    / ((2 / 2.354**2) * torch.pow(self.fwhm.value(mode), 2))
                    )
                )
                        

class Moffat(Ellip, LightProfile):
    def __init__(
    self, 
    cen_x: Parameters, 
    cen_y: Parameters, 
    pa: Parameters,
    axis_r: Parameters,
    
    mag: Parameters,
    con: Parameters,
    fwhm: Parameters,
    
    box: Parameters=None,
    ):
        """
        The Moffat profile. 

        Parameters
        ----------
        cen_x
            The galaxy center in x-xis in arc-second.
        cen_y
            The galaxy center in y-xis in arc-second.
        pa
            The position angle of the ellipse.
        axis_r
            The axis ratio of the ellipse.      
        eff_r
            The effective radius, which encompasses half of the total profile flux.
        mag
            The total magnitude of the galaxy.
        con
            The concentration index, dictates whether the distribution is 
            more Lorentzian-like (n =1) or Gaussian-like (n → ∞).
        fwhm
            The sersic index, which dictates the profile's curvature            
        box
            The boxness of the ellipse.  
        """  
        super().__init__(cen_x=cen_x, cen_y=cen_y, pa=pa, axis_r=axis_r, box=box)
        self.mag = mag
        self.con = con
        self.fwhm = fwhm
        
        
    def image_via_radii_from(self, radius, mode="updating_model"):
        """
        Returns the model image of the Moffat profile from radius.

        Parameters
        ----------
        radius
            The radius of the equivalent circle.
        mode
            Determine which kind of values will be called in parameter 
            object.
        """              
        Rd = (self.fwhm.value(mode) 
              / (2 * torch.pow(
                  torch.pow(
                      2, 1 / self.con.value(mode)
                      ) - 1, 0.5
                  )))
        I_e = (torch.pow(10, -0.4 * self.mag.value(mode))
                * self.R_box(mode)*(self.con.value(mode)-1)
               / (torch.pi * self.axis_r.value(mode) * torch.pow(Rd, 2))
               )
        
        return I_e * torch.pow(1 + torch.pow(radius / Rd, 2), 
                               -self.con.value(mode))
                        


class Sky(LightProfile):
    def __init__(
    self, 
    sky_bg: Parameters,
    ):  
        """
        The flat sky light profile. In defalt, this profile will not be
        convolved by PSF image.

        Parameters
        ----------
        sky_bg
            The sky mean value across all pixels without a radial matrix
             
        """     
        super().__init__()
        self.sky_bg = sky_bg
        self.psf = False

    def image_via_grid_from(self, 
                            grid: Union[torch.tensor, torch.tensor], 
                            mode="updating_model"):
        """
        Returns the model image of the flat sky light profile from a grid.

        Parameters
        ----------
        grid
            The coordinate grid (galaxy-centered).
        """        
        return torch.ones_like([0]) * self.sky_bg.value(mode)

   


                        



import numpy as np
import torch
from astropy.io import fits
from typing import List

from galmoss.profile.light_profile import LightProfile


class Profiles(): 
  def __init__(self, **kwargs):
    """
    This class stores the galaxy profiles that will be fitted.

    Parameters
    ----------
    **kwargs: 
        The galaxy profiles, e.g., profile1 = sersic
        
    Attributes
    ----------
    lightProfile_dict
        The dictionary stores all the defined galaxy profiles.
    no_conv_profile_dict
        The dictionary stores the profiles should not be convolved with PSF image.
    conv_profile_dict
        The dictionary stores the profiles should be convolved with PSF image.
    variableParam
        The dictionary stores all the parameters should be fitted.
    fixedParam
        The dictionary stores all the parameters should not be fitted.

    Note:
        The repeat parameters will appear only once in the Param dictionaries.
        For example, if two profiles share the galaxy center, the cen_x and cen_y will
        only appear once in dictionary they belong to.
    """     
    self.lightProfile_dict = {}
    self.no_conv_profile_dict = {}
    self.conv_profile_dict = {} 
    self.variableParam = {}
    self.fixedParam = {}
    unique_variable_param= set()
    unique_fixed_param = set()    
    for name, val in kwargs.items():
        if isinstance(val, LightProfile):
            self.lightProfile_dict[name] = val
            if getattr(val, 'psf', False):
                self.conv_profile_dict[name] = val
            else:
                self.no_conv_profile_dict[name] = val
            val.manage_params(name=name)
            unique_variable_param = unique_variable_param | set(val.variable_dic.values())
            unique_fixed_param = unique_fixed_param | set(val.fixed_dic.values())
        else:
            raise ValueError("The object be load in Profile class should be light profile!")    

    for param in unique_variable_param:
        count = sum(key.startswith(param.param_name) for key in self.variableParam.keys())
        new_key = param.param_name if count == 0 else f"{param.param_name}_{count}"
        self.variableParam[new_key] = param
    for param in unique_fixed_param:
        count = sum(key.startswith(param.param_name) for key in self.fixedParam.keys())
        new_key = param.param_name if count == 0 else f"{param.param_name}_{count}"
        self.fixedParam[new_key] = param
    setattr(self, name, val)


class DataSet():
    """
    Load the images and Parameters in parallel

    Parameters
    ----------
    func
        A function which returns the parameters need to be load in the optimism.

    Returns
    -------
        A function that returns the scale-parameters need to be load in the optimism.
    """
    def __init__(
        self, 
        galaxy_index: List[str],
        image_path: List[str],
        sigma_path: List[str],
        psf_path: List[str] = None,
        mask_path: List[str] = None,
        image_hdu: int = 0,  
        sigma_hdu: int = 0, 
        psf_hdu: int = 0,
        mask_hdu: int = 0,
        mask_index: List[str] = None,
        mask_mode = "bg",
        result_path = None,
        result_type = "FITS",
        img_block_path = None,
        data_type: torch.dtype = torch.float32,
        device: str = 'cuda'
    ):
        """
        This class stores the important attributes.

        Parameters
        ----------
        galaxy_index: 
            The list of the galaxy names.
        image/sigma/psf/mask_path
            The list of the galaxy/sigma/psf/mask image path.
        image/sigma/psf/mask_hdu
            The hdu number of the galaxy/sigma/psf/mask images.
        mask_index
            The list of the mask index, representing which region belongs
            to the correponding galaxy.
        mask_mode
            `bg`: The mask will only delete the region belongs to other 
            celestial bodies.
            `exclusive`: The mask will only remain the region belongs to 
            the corresponding galaxy.
        result_path
            The saving path of fitted value of parameters.
        result_type
            The data type of fitted value of parameters, should be FITS, 
            CSV or HDF5.
        img_block_path
            The saving path of fitted value of galaxy img_block, which includes
            galaxy image, model image, residual image and sub-component images.
        data_type
            The data type of matrix during calculations
        device
            Calculate the fitting process on CPU or GPU.

        Attributes
        ----------
        mode
            To change the dataset mode into fitting, bootstrap or covar_mat. 
        """     
        self.galaxy_index = [galaxy_index] if isinstance(galaxy_index, str) else galaxy_index 
        self.image_path = [image_path] if isinstance(image_path, str) else image_path 
        self.image_hdu = image_hdu

        self.psf_path = [psf_path] if isinstance(psf_path, str) else psf_path 
        self.psf_hdu = psf_hdu
        self.sigma_path = [sigma_path] if isinstance(sigma_path, str) else sigma_path 
        self.sigma_hdu = sigma_hdu
        self.mask_path = [mask_path] if isinstance(mask_path, str) else mask_path 
        self.mask_hdu = mask_hdu
        
        if mask_mode not in ["bg", "exclusive"]:
            raise ValueError(
                "Wrong mask_mode text, should be one of: bg, exclusive!"
            )   
        self.mask_mode = mask_mode

        self.mask_index = [mask_index] if isinstance(mask_index, int) else mask_index
        if ((mask_path is not None) and (self.mask_index is None)):
            self.mask_index = np.ones((len(self.galaxy_index)))
        
        if not (isinstance(result_path, str) or result_path is None):
            raise ValueError("The type of result_path should be str !")    
        self.result_path = result_path
        
        if result_type not in ["FITS", "CSV", "HDF5"]:
            raise ValueError("Result_type should be one of FITS, CSV or HDF5!")
        self.result_type = result_type

        if not (isinstance(img_block_path, str) or img_block_path is None):
            raise ValueError("The type of img_block_path should be str !")    
        self.img_block_path = img_block_path    
        self.data_type = data_type  
        self.device = device

        self.mode = "fitting"
        
    def read_data(self, 
                  path: str, 
                  hdu: int):
        """
        To read the data from fits file.
        """
        HDU = fits.open(path)
        return torch.from_numpy(HDU[hdu].data.astype(np.float64)).to(self.device)
  
    @property
    def data_size(self):
        """
        To get the data size from the first galaxy's image.
        """        
        HDU = fits.open(self.image_path[0])
        data = HDU[self.image_hdu].data.astype(np.float64)
        return np.shape(data)

    @classmethod
    def define_profiles(cls, *args, **kwargs):  
        """
        To define the profile class and set as an attribute in DataSet class.
        """              
        cls.Profiles = Profiles(**kwargs)

    def param_matrix_from_(self, 
                           param_dict: dict, 
                           mode: str, 
                           item: int):
        """
        To concat the param value matrix following the param_dict, mode and item.
        
        Parameters
        ----------
        param_dict
            The dictionary stores the keys (name of parameters) and values 
            (the parameter objects). These parameters need to be cancated 
            to a matrix that will be loaded into optimizer.
        mode
            The values of parameters will be called follow the mode in the 
            parameter objects, e.g., updating_model.
        item
            The index of values, which will be given automaticly when the 
            dataset is loaded in the data loader.    
        
        Returns
        -------
        total_matrix
            The finial matrix, being transformed to the given data type.            
        """   
        total_matrix = torch.empty((0,))
        for param in param_dict.values():
            if param.pix_scale is not None and hasattr(param, 'pix_scale_device')== False:
                param.pix_scale_device = torch.tensor(param.pix_scale).to(self.device).reshape(-1, 1)            
            if param.value(mode).shape[0] < len(self.galaxy_index):
                raise ValueError("The size of parameter doesn't match the"
                "number of galaxy, expect batch of galaxies with number of"
                "{} have a size of parameter of {}, but get {}".format(
                    len(self.galaxy_index), len(self.galaxy_index), 
                    param.value(mode).shape[1])) 
            total_matrix = torch.concat((total_matrix, param.value(mode)[item]), dim=0)          
        return total_matrix.to(self.data_type)

    def sigma_pretreatment(self, sigma_datadata: torch.tensor):
        """
        To reset the illegal sigma values.
        """         
        max_sigma = torch.max(sigma_datadata)
        sigma_datadata[sigma_datadata <= 0] = max_sigma  
        return sigma_datadata 
    
    def __getitem__(self, item: int):
        """
        Output the data once the DataSet being loaded by data loader.
        """ 
        self.data_cube_idx = ["data", "sigma"]
        fits_data = self.read_data(self.image_path[item], self.image_hdu).to(self.data_type)
        sigma_data = self.sigma_pretreatment(self.read_data(self.sigma_path[item], self.sigma_hdu)).to(self.data_type)

        data_matrix = torch.cat([fits_data.unsqueeze(0), sigma_data.unsqueeze(0)], dim=0)  
        if self.psf_path is not None:
            psf_data = self.read_data(self.psf_path[item], self.psf_hdu).to(self.data_type)
        else:
            psf_data = torch.zeros((1)).to(self.device)

        if self.mask_path is not None:   
            mask_data = self.read_data(self.mask_path[item], self.mask_hdu).to(self.data_type)
            # if self.mask_index is not None:
            if self.mask_mode == "bg":
                mask_data[(mask_data != self.mask_index[item]) & (mask_data != 0)] = 999
                mask_data[mask_data != 999] = 1
                mask_data[mask_data == 999] = 0
            else:
                mask_data[mask_data != int(self.mask_index[item])] = 0
                mask_data[mask_data == int(self.mask_index[item])] = 1
            data_matrix = torch.cat([data_matrix, mask_data.unsqueeze(0)], dim=0)
            self.data_cube_idx.append("seg")
        
        if self.mode == "fitting":
            return (
                self.galaxy_index[item], 
                data_matrix, 
                psf_data, 
                self.param_matrix_from_(
                    self.Profiles.variableParam, 
                    "load_initial_in_optim_list", 
                    item).to(self.device), 
                self.param_matrix_from_(
                    self.Profiles.fixedParam, 
                    "load_initial_in_optim_list", 
                    item).to(self.device)
            )
        elif self.mode == "bootstrap":         
            return (
                self.galaxy_index[item], 
                data_matrix, 
                psf_data, 
                self.param_matrix_from_(
                    self.Profiles.bsp_variableParam, 
                    "load_best_in_optim_list", 
                    item).to(self.device), 
                self.param_matrix_from_(
                    self.Profiles.bsp_fixedParam, 
                    "load_initial_in_optim_list", 
                    item).to(self.device)
                )
        elif self.mode == "covar_mat":
            return (
                self.galaxy_index[item], 
                data_matrix, 
                self.param_matrix_from_(
                    self.Profiles.variableParam, 
                    "load_best_in_jacobian", 
                    item).to(self.device), 
                self.param_matrix_from_(
                    self.Profiles.fixedParam, 
                    "load_initial_in_optim_list", 
                    item).to(self.device)
                )
        else:
            assert "Unavailable"
        
    def __len__(self):
        return len(self.galaxy_index)    
  
  
  
    


import numpy as np
import torch
import torch.utils.data as Data
from torch.func import vmap, jacrev
from tqdm import tqdm


class Bootstrap:
    """
    Class for uncertainty calculation method Bootstrap.
    """ 
    def prepair_data(self, center_fix: bool):
        """
        Refitting in bootstrap process often meets challenge if the
        pixels near the galaxy center is not be resampled. This function 
        is for changing the fix statement of center_x and center_y. If 
        center_fix == True, the profile center will be fixed, which will
        help to make the refitting process more stable.

        Parameters
        ----------
        center_fix
            Bool augument, to determin the center parameters will be fixed
            or not.
        """
        if center_fix == True:
            self.dataset.Profiles.bsp_variableParam = {}
            self.dataset.Profiles.bsp_fixedParam = {}
            for param_name, param in self.dataset.Profiles.variableParam.items():
                key = param.param_name
                if key == "cen_x" or key == "cen_y":
                    param.fit = False
                    self.dataset.Profiles.bsp_fixedParam[param_name] = param
                else:
                    self.dataset.Profiles.bsp_variableParam[param_name] = param  
            for param in self.dataset.Profiles.fixedParam.values():
                self.dataset.Profiles.bsp_fixedParam[param_name] = param    

    def resample_mask(self, mask_data: torch.tensor):
        """
        this function is for resampling the pixels of galaxy image data. 
        
        Parameters
        ----------        
        mask_data:
            The mask data will be resampled.
        
        Returns
        -------
            The resampled mask data.
        """
        random_integers = torch.randint(0, 
                                        np.prod(mask_data.shape[1:]), 
                                        (np.prod(mask_data.shape[1:]), )
                                        )           
        flat_mask = mask_data.reshape(len(self.index), -1)
        new_mask = torch.zeros(np.shape(flat_mask)).to(self.dataset.device)
        unique_arr, counts = torch.unique(random_integers, return_counts=True)
        new_mask[:, unique_arr] = (flat_mask[:, unique_arr] 
                                   * counts.to(self.dataset.device))
        new_mask = new_mask.reshape(len(self.index), 
                                    np.shape(mask_data)[1], 
                                    np.shape(mask_data)[2])     
        return new_mask
             
    def calculate_uncertainty(self, P_dict):
        """
        Calculate the fitting uncertainty via calculating the variance 
        among the fitted value of parameters.
        """
        for param in P_dict.values():
            uncertainty = torch.std(param.bsp_best_value, dim=2)

            if param.dim == "1d":
                param.uncertainty = param.inv_from_mapping_uc(
                    uncertainty.reshape(param.param_length))
            
    def bootstrap(self, 
                  n_resample: int = 100, 
                  iteration: int = 100, 
                  center_fix: bool = True
        ):
        """
        Refit the resampled galaxies, as a way to calculating fitting 
        uncertianties.

        Parameters
        ----------
        n_resample
            The time of the resampling process. The suggest value is 100.
        iteration
            The maximum fitting iteration, the fitting process will be 
            shut down when reaching this number.
        center_fix
            To decide whether set the fit statement of paramter 
            (Center_x & Center_y) to False. The suggest value is True. 
        """
        self.dataset.mode = "bootstrap"
        self.prepair_data(center_fix)
        bsp_dataset = Data.DataLoader(dataset=self.dataset, 
                                      batch_size=self.batch_size, 
                                      shuffle=False)
        for (self.index, 
             self.data_cube, 
             self.psf, 
             param_final_value, 
             consant_value) in bsp_dataset:

            for _ in tqdm(np.arange(n_resample)):              
                optim_list = self.make_optim_list(
                    self.dataset.Profiles.bsp_variableParam, 
                    param_final_value)
                self.load_constant(self.dataset.Profiles.bsp_fixedParam, 
                                   consant_value)
                optimizer = self.optimizer(optim_list) 
                fitting_metrics = self.initialize_fitting_metrics()
                
                for iter in np.arange(iteration): 
                    self.refresh_variable(
                        self.dataset.Profiles.bsp_variableParam, 
                        optim_list)
                
                    conv_model = self.modeling(
                        self.dataset.Profiles.conv_profile_dict
                        )
                    no_conv_model = self.modeling(
                        self.dataset.Profiles.no_conv_profile_dict
                        )
                    model_tot = (self.add_convolution(conv_model.sum(axis=0)) 
                                 + no_conv_model.sum(axis=0))
                    
                    loss, chi_mu = self.optimizing(self.data_cube[:, 0], 
                                                    model_tot, 
                                                    self.data_cube[:, 1])  
                    loss.backward()
                    optimizer.step()  
                    optimizer.zero_grad()    

                    update, fitting_metrics = self.detect_update(
                        fitting_metrics,
                        chi_mu, 
                        iter
                    )  
                    if not update:
                        break    
                self.refitted_param_append(
                    self.dataset.Profiles.bsp_variableParam, 
                    sampling_idx = _
                    )


            self.refitted_param_append(self.dataset.Profiles.bsp_variableParam, 
                                      total=True)  
        self.calculate_uncertainty(self.dataset.Profiles.bsp_variableParam)

    def refitted_param_append(self, 
                            P_dict: dict, 
                            total:bool = False,
                            sampling_idx: int = 999):
        """
        This function is for storing the parameters refitting value inside 
        bootstrapping process. If total is True, which means the 
        process is appending batch refitted value to total refitted value.
        If total is False, which means the process is appending total 
        refitted value with the previous total refitted value (in principle
        each galaxy's each parameters will have n_sampled values)
        
        Parameters
        ----------
        sampling_idx
            The idx of sampling process, if zero, need to reset bsp_batch_best_value
        P_dict
            Dictionary of variable parameters
        total
            Determines to running which kind of appending process.
        """
        if not total:
            for param in P_dict.values():
                if (hasattr(param, 'bsp_batch_best_value') and sampling_idx != 0):
                    param.bsp_batch_best_value = torch.cat((
                        param.bsp_batch_best_value, 
                        param.batched_mapped_best_value.unsqueeze(2)), 
                        dim=2)     
                else:
                    param.bsp_batch_best_value = (param
                                                   .batched_mapped_best_value
                                                   .unsqueeze(2))      
        else:
            for param in P_dict.values():              
                if hasattr(param, 'bsp_best_value'):
                    param.bsp_best_value = torch.cat((
                        param.bsp_best_value, 
                        param.bsp_batch_best_value), 
                        dim=0)
                else:
                    param.bsp_best_value = param.bsp_batch_best_value


class CovarianceMatrix:
    """
    Class for uncertainty calculation method CovarianceMatrix.
    """    
    def CM_uncertainty_append(self, batched_uncertainty: torch.tensor):
        """
        Append the uncertainty values for each batch of galaxies after
        the completion of each uncertainty calculation process.

        Parameters
        ----------
        batched_uncertainty
            The uncertainty for each batch of galaxies.
        """          
        if hasattr(self, 'total_uncertainty'):
            self.total_uncertainty = torch.concat((self.total_uncertainty, 
                                                   batched_uncertainty), 
                                                  dim=0)                 
        else:
            self.total_uncertainty = batched_uncertainty                    
        
    def calculate_uncertainty_from_J(self, jacobian_matrix: torch.tensor):
        """
        Use Jacobian matrix to appraxmate Hessian matrix, then use the 
        diagnose of the inverse of Hessian matrix to calculate parameter 
        fitting uncertainty.

        Parameters
        ----------
        jacobian_matrix
            The corresponding jacobian matrix.

        Returns
        -------
        uncertainty
            The uncertainty calculated from Jacobian matrix.
        """           
        new_J, filtered_eyed_W = self.reconstruct(jacobian_matrix)
        jTj = torch.bmm(torch.bmm(new_J, filtered_eyed_W), 
                        new_J.transpose(1, 2))   
        LU, pivots = torch.linalg.lu_factor(jTj.detach())
        
        jTj_inv = torch.linalg.lu_solve(
            LU, pivots, 
            torch.eye(len(self.dataset.Profiles.variableParam))
            .to(self.dataset.device)
            .repeat(len(self.index), 1, 1)) 
        
        uncertainty = torch.pow(torch.diagonal(jTj_inv, dim1=-2, dim2=-1), 0.5)

        return uncertainty
            
    def store_uncertainty(self, param_value):
        """
        This function is responsible for separating and storing the 
        uncertainties associated with each fitting parameter in the 
        corresponding parameter object. It iterates through each parameter 
        involved in the fitting process, assigning the relevant uncertainty 
        values from the provided tensor.
        """
        for idx, param in enumerate(self.dataset.Profiles.variableParam
                                    .values()):
            param.uncertainty = param.inv_from_mapping_uc(self.total_uncertainty[:, idx])
                              
    def reconstruct(self, J:torch.Tensor) -> [torch.Tensor, torch.Tensor]:
        """
        To extract the effective value from weight matrix and Jacobian 
        matrix based on the mask matrix, and resize to flat matrix.

        Parameters
        ----------
        J
            The original Jacobian matrix.
        Returns
        -------
        new_J
            The effective and flat Jacobian matrix.
        new_eyed_W
            The eye matrix, with values equals flat weight matrix.
        """
        flat_weight = ((1 / (torch.pow(self.data_cube[:, 1] , 2)))
                           .view(len(self.index), 1, -1))     
        if self.dataset.mask_path is not None:   
            reshape_mask = self.data_cube[:, 2].reshape(len(self.index), -1)
            eff_pix_lengths = torch.sum(reshape_mask, dim=1).int()

            new_J = torch.zeros(len(self.index), 
                                len(self.dataset.Profiles.variableParam), 
                                torch.sum(reshape_mask, dim=1).int().max(), 
                                dtype=torch.float32).to(self.dataset.device) 
            new_W = torch.zeros(len(self.index), 
                                1, 
                                torch.sum(reshape_mask, dim=1).int().max(), 
                                dtype=torch.float32).to(self.dataset.device)  

            for i in range(len(self.index)): 
                eff_counts = eff_pix_lengths[i].item()
                where_eff = torch.nonzero(reshape_mask[i] == 1, 
                                                   as_tuple=True)[0]
                new_W[i, 0, : eff_counts] = flat_weight[i, :, where_eff]
                new_J[i, :, : eff_counts] = J[i, :, where_eff]
            eye_size = int(torch.sum(reshape_mask, dim=1).max().item())        
                                    
        else:
            new_J = J
            new_W = flat_weight            
            eye_size = np.prod(self.data_cube[:, 1].shape[1:])
        
        eyes_mask = (torch.eye(eye_size)
                     .expand(len(new_J), 
                             eye_size, 
                             eye_size).to(self.dataset.device))
        new_eyed_W = eyes_mask * new_W.float() 
        
        return new_J, new_eyed_W
         
    def jacobian(self, param: torch.Tensor) -> torch.Tensor:
        """
        This function **would not** be used to calculate model value, 
        only to be load in jacrevm for calculating the Jacobian matrix. 
        
        Parameters
        ----------
        param
            The fitted parameters after fitting process.
        model
            The model image calculated by fitted parameters.
        """
        ams = 0
        for param_class in self.dataset.Profiles.variableParam.values():
            param_class.batched_updating_value = (
                param[ams: ams + param_class.dim2_length])
            ams += param_class.dim2_length
                
        conv_model = self.modeling(self.dataset.Profiles.conv_profile_dict, 
                                   mode="updating_model") 
        no_conv_model = self.modeling(self.dataset.Profiles.no_conv_profile_dict,
                                      mode="updating_model")    
        
        model_tot =( conv_model.sum(axis=0) + no_conv_model.sum(axis=0)).sum(axis=0)
        model = model_tot.view((self.dataset.data_size[0] 
                                * self.dataset.data_size[1]))
        return model        
       
    def covar_mat(self, bs: int = 2):
        """
        Calculate the fitting uncertainty for each parameters.

        Parameters
        ----------
        bs
            the batch-size for calculating covariance matrix. 
            Cause the huge calculation cost in Jacobian matrix,
            the bs here should be much smaller than the fitting
            batch size.
        """
        self.dataset.mode = "covar_mat"
        dataset = Data.DataLoader(dataset=self.dataset, 
                                  batch_size=bs, 
                                  shuffle=False)
    

        for (self.index, 
             self.data_cube, 
             param_final_value, 
             consant_value) in dataset:
            self.load_constant(self.dataset.Profiles.fixedParam, consant_value)
            with torch.no_grad():
                jacobian = vmap(jacrev(self.jacobian, argnums=0), 
                                in_dims=0)(param_final_value)
                J = torch.transpose(jacobian, 1, 2)
            self.CM_uncertainty_append(self.calculate_uncertainty_from_J(J))

        self.store_uncertainty(param_final_value)


class Uncertainty(Bootstrap, CovarianceMatrix):
    def __init__(self):  
        super().__init__() 

    def uncertainty(self, method = "covar_mat", *args, **kwargs):
        """
        Start calculating fitting uncertainty follows the method, and 
        resave the fitted value.
        """             
        if next(
            iter(self.dataset.Profiles.variableParam.values())
            ).total_mapped_best_value is not None: 
            if method == "covar_mat":
                self.covar_mat(*args, **kwargs)
            elif method == "bootstrap":
                self.bootstrap(*args, **kwargs)
            else:
                raise ValueError("Wrong method text, should be one of: "
                                 "covar_mat or bstrap")
        else:
            raise ValueError("There isn't any fitted value bein stored inside"
                             " each parameter object, please refit, or reload"
                             " the fitted value into the parameter object, and"
                             " then change the iteration number to zero!")

        if isinstance(self.dataset.result_path, str):
                self.save_param()   
                print("The parameter fitting result is resaved in {} "
                      "successfully with uncertainty calculated by {}!"
                      .format(self.dataset.img_block_path, method))  
        else:
            print("The parameter fitting result becuase result_path is not set.")   
 




from setuptools import setup, find_packages

with open('README.rst', 'r',encoding='utf-8') as fp:
    long_description = fp.read()

setup(
    name='galmoss',
    version='2.0.7',
    description='A Python-based, Torch-powered tool for two-dimensional fitting of galaxy profiles. By seamlessly enabling GPU parallelization, GalMOSS meets the high computational demands of large-scale galaxy surveys.',
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url='https://github.com/Chenmi0619/GALMoss',
    keywords=['Astronomy data analysis', 'Astronomy toolbox', 'Galaxy profile fitting'],
    author='Chen Mi',
    author_email='chenmiastro@gmail.com',
    classifiers=[
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3',
    ],
    python_requires=">=3.9",
    packages=find_packages(),
)




import numpy as np
import torch
from itertools import chain
import pandas as pd
from tqdm import tqdm
import torch_optimizer as optim
import os
import h5py
import torch.utils.data as Data
import torch.nn.functional as F
from astropy.io import fits
from astropy.table import Table
from typing import Union

from galmoss.profile.light_profile import LightProfile
from galmoss.data import DataSet
from galmoss.uncertainty import Uncertainty

     
class FittingRepo():
    """
    This class owns functions that relates to fitting metrics initializing, 
    updating and storaging in fitting process. 
    """
    
    def initialize_fitting_metrics(self):
        """
        Initial the matrix stores fitting metrics.
        
        Returns
        -------
        stop_update_count
            The number of galaxies that already stop fitting.
        batched_eff_iter
            The iteration being used when the belonging galaxy stop fitting 
            when reaching the updating shreshold in this batch.
        batched_best_chi_mu
            The best chi_mu value of belonging galaxy in this batch. 
        """
        stop_update_count = torch.zeros((len(self._index))).to(self.dataset.device)
        batched_eff_iter = torch.ones((len(self._index))).to(self.dataset.device)
        batched_best_chi_mu = torch.empty((0,)).to(self.dataset.device)
        return (stop_update_count, batched_eff_iter, batched_best_chi_mu)
         
    # def update_best_chi_mu(self, 
    #                         chi_mu: torch.tensor, 
    #                         index: Union[torch.tensor, bool] = None):
    #     """
    #     Update the best chi_mu value during the optimization process. Not 
    #     every iteration in the optimization process results in a better 
    #     outcome with a lower chi-square. Therefore, updates are made only 
    #     in iterations that successfully reduce the chi-square, guided by 
    #     the given index.

    #     Parameters
    #     ----------
    #     chi_mu
    #         The chi_mu being calculated in one fitting iteration.
    #     index
    #         Specifies the position in the best chi_mu value matrix to be
    #         updated. If None, updated the entire matrix.
    #     """            
    #     if index == None:
    #         self.batched_best_chi_mu = chi_mu
    #     else:
    #         self.batched_best_chi_mu[index] = chi_mu[index]
                    
    def fitting_metrics_append(self, batched_eff_iter, batched_best_chi_mu):
        """
        Append the fitting_metrics for each batch of galaxies after
        the completion of each batch fitting process.

        Parameters
        ----------
        batched_eff_iter
            The iteration being used when the belonging galaxy stop fitting 
            when reaching the updating shreshold in this batch.
        batched_best_chi_mu
            The best chi_mu value of belonging galaxy in this batch. 
        """   
        if hasattr(self, 'total_best_chi_mu'):
            self.total_best_chi_mu = torch.concat(
                (self.total_best_chi_mu, batched_best_chi_mu), dim=0)        
        else:
            self.total_best_chi_mu = batched_best_chi_mu
        
        if hasattr(self, 'total_eff_iter'):
            self.total_eff_iter = torch.concat(
                (self.total_eff_iter, batched_eff_iter), dim=0)        
        else:
            self.total_eff_iter = batched_eff_iter

    @property
    def chi_mu_numpy(self) -> np.ndarray:
        """
        Turns the chi_mu data after fitting into numpy.

        Returns
        -------
        np.ndarray
            The chi_mu values as a NumPy array.
        """
        return self._chi_mu.detach().cpu().numpy()

    def refresh_variable(self, P_dict, optim_list):
        """
        This function is for refresh the stepped value inside optimizer to
        the parameters instance, in sequence of the parameter diction 
        being given.

        Parameters
        ----------
        P_dict:
            The dictionary of variable parameters.
        optim_list:
            The optim list stores the stepped value of such parameters.
        """
        for ams, param in enumerate(P_dict.values()):
            data = optim_list[ams]['params'][0]
            param.update_updating_val(data)
 
    def load_constant(self, P_dict, P_value):
        """
        This function is for set the fixed parameters (constant) before 
        every batch-size of training.

        Parameters
        ----------
        P_dict:
            The dictionary of fixed parameters.
        P_value:
            The total value matrix stores the initial mapped value of such
            parameters.
        """        
        ams = 0
        for param in P_dict.values():
            param.update_updating_val(P_value[:, ams: ams+param.dim2_length])
            param.update_best_val(index=None)    
            ams += param.dim2_length    
  
    def update_param_best_value(self, index):
        """
        Update the best value of each parameters during the optimization
        process. 
        
        Parameters
        ----------
        index
            Specifies the position in the best value matrix to be
            refreshed. If None, refreshes the entire matrix.
        """          
        with torch.no_grad(): 
            for param in self.dataset.Profiles.variableParam.values():
                param.update_best_val(index)
                param.grad = 0
        
    def append_param_best_value(self):
        """
        Append the best fitted values of each parameters after the 
        completion of each batch fitting process.
        """ 
        for param in self.dataset.Profiles.variableParam.values():
            param.best_value_append(param.batched_mapped_best_value)
            
    def save_param(self): 
        """
        Save the best fitted values of each parameters after the total 
        fitting process ends. A dictionary will be used to store the key
        ("{}_{}_(varaible/fixed)".format(profile_name, param_name)) and 
        the corresponding fitted value.
        
        If the mode is not fitting, this function will save the 
        uncertainty together, with the key:
            ("{}_{}_err".format(profile_name, param_name))
        and the corresponding uncertainty value.
        
        After the dictionary is defined, the file will be saved follows 
        the attributes `result_type`.
        """         
        data = {}
        data["galaxy_idx"] = self.dataset.galaxy_index
        data["chi_mu"] = self.total_best_chi_mu.detach().cpu().numpy()
        for profile_name, profile in self.dataset.Profiles.lightProfile_dict.items():          
            for variable_param in profile.variable_dic.values():
                key = "{}_{}_(varaible)".format(
                    profile_name, 
                    variable_param.param_name
                    )
                if variable_param.dim == "1d":
                    data[key] = (variable_param.total_best_value_numpy
                                 .reshape(variable_param.param_length))
                if ((self.dataset.mode in ["bootstrap", "covar_mat"])
                    and (variable_param.fit)):
                    key = "{}_{}_err".format(
                        profile_name, 
                        variable_param.param_name
                        )
                    data[key] = variable_param.total_uncertainty_numpy  
            for fixed_param in profile.fixed_dic.values():
                key = "{}_{}_(fixed)".format(
                    profile_name, 
                    fixed_param.param_name
                    )
                data[key] = (fixed_param.total_initial_value_numpy
                                 .reshape(fixed_param.param_length)) 
        if self.dataset.result_type == "FITS":              
            hdu = fits.BinTableHDU(Table(data))
            hdu.writeto(os.path.join(self.dataset.result_path, 'result.fits'), 
                        overwrite=True) 
        elif self.dataset.result_type == "CSV":
            df = pd.DataFrame(data)
            df.to_csv(os.path.join(self.dataset.result_path, 'result.csv'), 
                      index=False)   
        else:
            with h5py.File(os.path.join(self.dataset.result_path, 
                                        'result.hdf5'), 'w') as hdf:
                for key, value in data.items():
                    hdf.create_dataset(key, data=value)              

    def save_image(self):
        """
        This method saves the img block as FITS file, which includes galaxy 
        image, model image, residual image and the image of sub-components. 
        Each sub-components will be convoluted either.
        """                 
        conv_model = self.modeling(
            profile_dict=self.dataset.Profiles.conv_profile_dict, 
            mode="best_model")
        no_conv_model = self.modeling(
            profile_dict=self.dataset.Profiles.no_conv_profile_dict, 
            mode="best_model")
        component_cube = torch.cat((conv_model, no_conv_model), dim=0)
                
        if self.dataset.psf_path is not None:
            model_tot = (self.add_convolution(conv_model.sum(axis=0)) 
                           + no_conv_model.sum(axis=0))
            residual = (self.data_cube[:, 0] - model_tot).detach().cpu().numpy()
            for profile_idx in np.arange(
                len(self.dataset.Profiles.conv_profile_dict)
                ):
                conv_model[profile_idx] = torch.squeeze(
                    F.conv2d(
                        conv_model[profile_idx].unsqueeze(0), 
                        self._psf.unsqueeze(1), 
                        padding="same", 
                        groups=len(self._index)
                        )
                ) 
        else:
            model_tot = conv_model.sum(axis=0) + no_conv_model.sum(axis=0)
            residual = (self.data_cube[:, 0] - model_tot).detach().cpu().numpy()
            for profile_idx in np.arange(
                len(self.dataset.Profiles.conv_profile_dict)
                ):
                conv_model[profile_idx] = conv_model[profile_idx]
                                     
        galaxy_data = self.data_cube[:, 0].detach().cpu().numpy()
        model_tot = model_tot.detach().cpu().numpy()
        for ams, galaxy in enumerate(self._index):
            primary_hdu = fits.PrimaryHDU()
            image_hdu = fits.ImageHDU(galaxy_data[ams])
            image_hdu.header['KEYWORD'] = 'data'

            model_hdu = fits.ImageHDU(model_tot[ams])
            model_hdu.header['KEYWORD'] = 'model'
                
            residual_hdu = fits.ImageHDU(residual[ams])
            residual_hdu.header['KEYWORD'] = 'residual'

            component_hdulist = [primary_hdu, image_hdu, model_hdu, residual_hdu]

            for abs, (key, key_value) in enumerate(
                chain(self.dataset.Profiles.conv_profile_dict.items(),
                self.dataset.Profiles.no_conv_profile_dict.items())
                ):  
                if key_value.saving:     
                    component_hdu = fits.ImageHDU(
                        component_cube[abs, ams].detach().cpu().numpy())
                    component_hdu.header['KEYWORD'] = key
                    component_hdulist.append(component_hdu)      
        
            component_hdulist = fits.HDUList(component_hdulist)
            component_hdulist.writeto(
                os.path.join(self.dataset.img_block_path, galaxy + '.fits'),
                overwrite=True
                )

          
class Fitting(FittingRepo, Uncertainty):
    """
    This class owns functions that relates to fitting process. 
    
    Parameters
    ----------
    dataset
        The object instanced from `DataSet` class, used for Load the 
        images and Parameters in parallel.
    batch_size
        The number of galaxies being fitted together in one batch. This
        number should as big as possible until the memory usage and GPU
        usage reach the limit.    
    iteration
        The number of fitting iteration for the fitting process, the 
        recommand number is 1,000 ~ 1,500.
    optimizer
        The type of optimizer. Galmoss supports any kind of optimizers in 
        PyTorch. The defalt one is DiffGrad.
    threshold & early_stop
        The galaxy which 
        (best_chi_mu  - chi_mu) / best_chi_mu  < `threshold` 
        more than `early_stop` time will stop fitting. The defalt number is
        10e-5 and 10.
    """
    def __init__(
        self, 
        dataset: DataSet,
        batch_size: int,
        iteration: int,
        optimizer = optim.DiffGrad,
        threshold = 1e-5,
        early_stop = 50
    ):
        self.dataset = dataset
        self.optimizer = optimizer
        
        if isinstance(batch_size, int) and batch_size > 0:
            self.batch_size = batch_size
        else:
            raise ValueError("batch size must be an integer greater than 0")

        if isinstance(iteration, int) and iteration > 0:
            self.iteration = iteration
        else:
            raise ValueError("iteration must be an integer greater than 0")

        if isinstance(threshold, float) and threshold > 0:
            self.threshold = threshold
        else:
            raise ValueError("threshold must be a float greater than 0")  

        if isinstance(early_stop, int) and early_stop > 0:
            self.early_stop = early_stop
        else:
            raise ValueError("early_stop must be an integer greater than 0")
                       
        super().__init__()  

    def make_optim_list(self, P_dict, P_value): 
        """
        Make the dictionary for being load in the optimizer.
        
        Parameters
        ----------
        P_dict
            The dictionary of parameter objects.
        P_value
            The total value matrix for the parameters inside P_dict.

        Returns
        -------
        optim_list
            The optimizer list, which includes dictionaries for each 
            parameters, which contains values and learning rate.
        """          
        optim_list = []
        ams = 0
        for param in P_dict.values():        
            param_dict = {}
            param_value = P_value[:, ams: ams+param.dim2_length].clone()
            param_dict['params'] = param_value.detach().requires_grad_(True)
            param_dict['lr'] = param.step_length
            optim_list.append(param_dict)
            ams += param.dim2_length
        return optim_list
                  
    def modeling(self, profile_dict, mode="updating_model"):
        """
        Generate the total model image matrix.
        
        Parameters
        ----------
        profile_dict
            The dictionary of profile objects.
        mode
            The string representing which value to choose from the 
            parameters.

        Returns
        -------
        model_matrix
            The total model matrix, with shape 
            (profile_num, galaxy_num, n, m).
            (n, m) is the data shape.
        """            
        model_repo = []
        for profile in profile_dict.values():
            model_repo.append(profile.image_via_grid_from(self.grid, mode))
        if model_repo:
            model_matrix = torch.stack(model_repo) 
        else:
            model_matrix = torch.zeros((0)).to(self.dataset.device)
        return model_matrix
      
    def make_grid(self):
        """
        Generate the grid for fitting process.
        """             
        x = torch.linspace(0.5, 
                           self.dataset.data_size[1]-0.5, 
                           self.dataset.data_size[1], 
                           dtype=self.dataset.data_type).to(self.dataset.device)
        y = torch.linspace(self.dataset.data_size[0]-0.5, 
                           0.5, 
                           self.dataset.data_size[0], 
                           dtype=self.dataset.data_type).to(self.dataset.device)
        xy = torch.meshgrid(y, x, indexing='ij')      
        self.grid = xy
          
    def revise_mask(self):
        """
        To generage the mask image follows various mode.
        If the fitting process already loads mask image, then the original 
        mask image will be call from `cls.data_cube`. Otherwise, it will 
        be a tensor filled with the scalar value 1, with the shape same as 
        the galaxy data image (In practical, the calculation will multiply 1 
        instead). The original mask image will be used In the fitting mode. 
        
        In bootstrap mode, the original mask data will be resampled.
        
        Returns
        -------
        mask
            The mask image that will be used in chi-square calculations.
        mask_num
            The effective pixel number of mask image, which will be used in
            freedom degree calculations.
        """

        if "seg" in self.dataset.data_cube_idx:
            if self.dataset.mode == "bootstrap":
                mask = self.resample_mask(
                    self.data_cube[:, 
                                   self.dataset.data_cube_idx.index("seg")
                                   ]
                    )
                mask_num = torch.sum(mask, dim=[1, 2])
            else:
                mask = self.data_cube[:, 
                                      self.dataset.data_cube_idx.index("seg")
                                      ]
                mask_num = torch.sum(mask, dim=[1, 2])                
        else:
            if self.dataset.mode == "bootstrap":
                mask = self.resample_mask(
                    torch.ones(
                        self.data_cube[:, 
                                       self.dataset.data_cube_idx.index("data")
                                       ].shape
                    ).to(self.dataset.device)
                )
                mask_num = torch.sum(mask, dim=[1, 2])
            else:
                mask = 1
                mask_num = self.dataset.data_size[0] * self.dataset.data_size[1]
        return mask, mask_num

    def add_convolution(self, to_conv_model):
        """
        Make the convolution to the input model image with the
        psf images.

        Parameters
        ----------
        to_conv_model
            The images need to be convolved.

        Returns
        -------
            The convolved images.
        """      
        if (self.dataset.psf_path is not None) and (len(to_conv_model) >0):
            noramlize_factor = torch.sum(
                self._psf, 
                dim=[1, 2]
                ).unsqueeze(1).unsqueeze(1)
            conved_model = torch.squeeze(F.conv2d(
                to_conv_model.unsqueeze(0), 
                self._psf.unsqueeze(1), 
                padding="same", 
                groups=len(self._index))) / noramlize_factor        
        else:
            conved_model = to_conv_model
        return conved_model
       
    def detect_update(self, fitting_metrics, chi_mu, iter):
        """
        Not every iteration in the optimization process results in a 
        better outcome with a lower chi-square. Therefore, updates are 
        made only in iterations that successfully reduce the chi-square, 
        which are detected here in `is_better`.

        `improvement_small` is galaxies their Δchi_mu is smaller than the 
        thresholds. if `early_stop` number of iterations achive this statement 
        continuously, this galaxy will be marked as early-stop galaxy.
        At present, only all the galaxies are early-stoped galaxy, the fitting 
        process will finish.
        
        Parameters
        ----------
        fitting_metrics
            A tuple includes stop_update_count, batched_eff_iter and
            batched_best_chi_mu. `stop_update_count` represents the
            number of iterations achive `improvement_small` statement 
            continuously. `batched_eff_iter` represents how much 
            iterations each galaxy used to achive the best value.
            `batched_best_chi_mu` represents the best chi_mu.
        chi_mu
            The chi_mu being calculated in this iteration.
        iter
            The value of this iteration.
            
        Returns
        -------
        update
            If True, continue fitting, otherwise break the loop.
        fitting_metrics 
            The updated fitting_metrics.
        """          
        (stop_update_count, 
         batched_eff_iter, 
         batched_best_chi_mu) = fitting_metrics
        
        if iter == 0:
            batched_best_chi_mu = chi_mu
            stop_update_count += 1
            self.update_param_best_value(index=None)
           
        else: 
            is_better = chi_mu < batched_best_chi_mu
            improvement_small = ((torch.abs(chi_mu - 1) 
                                 < torch.abs(batched_best_chi_mu - 1)) 
                                 & ((batched_best_chi_mu - chi_mu) 
                                    / batched_best_chi_mu  < self.threshold))
            
            batched_eff_iter[is_better] = iter
            batched_best_chi_mu[is_better] = chi_mu[is_better]            
            stop_update_count[~improvement_small] = 0
            stop_update_count[improvement_small] += 1 
            self.update_param_best_value(is_better)
       
            
        if torch.any(stop_update_count >= self.early_stop):
            num_early_stop = torch.sum(stop_update_count >= self.early_stop).item()  
        else:
            num_early_stop = 0

        update = (num_early_stop != len(self._index))
        return  update, (stop_update_count, batched_eff_iter, batched_best_chi_mu)     
  
    def optimizing(self, data, model, sigma): 
        """
        Calculate the residual for this fitting iteration.

        Parameters
        ----------
        data
            The galaxy image data in this batch.
        model
            The galaxy model data in this batch.
        sigma
            The galaxy sigma data in this batch.

        Returns
        -------
        total_residual
            The total residual in this fitting iteration, acts as float.
        chi_mu 
            The residual (acts as chi_mu) for each galaxy.
        """          
        mask, mask_num = self.revise_mask()
        chi = torch.pow((data - model)*mask, 2) / torch.pow(sigma, 2)
        if torch.isnan(model).any():
            nan_mask = torch.isnan(model).any(dim=2).any(dim=1)
            nan_indices = torch.where(nan_mask)
            chi[nan_indices] = 0   
        chi_mu = (torch.sum(chi, dim=[1, 2]) 
                   / (mask_num - len(self.dataset.Profiles.variableParam)))      
        total_residual = torch.sum(chi)
        return total_residual, chi_mu
  
    def fit(self):
        """
        The fitting process.

        After setting the dataset mode in `fitting`, we load it into the 
        data loader.

        Then we make the grid for model generation.

        The galaxies are loaded in batch using the data loader, and in 
        each batch we initialize the fitting metrics, fixed parameter
        values and optim_list.

        When each batch finishes fitting, their img_block will be saved 
        directly. When the total fitting process is finished, the fitted 
        parameter values will be saved in the end.
        """          
        self.dataset.mode = "fitting"
        train_dataset = Data.DataLoader(dataset=self.dataset, 
                                        batch_size=self.batch_size, 
                                        shuffle=False)
        self.make_grid()
        
        for (self._index, 
             self.data_cube, 
             self._psf, 
             _param_final_value,
             _consant_value) in train_dataset:
            
            fitting_metrics = self.initialize_fitting_metrics()
            self.load_constant(self.dataset.Profiles.fixedParam, 
                               _consant_value)
            optim_list = self.make_optim_list(
                self.dataset.Profiles.variableParam,
                _param_final_value
                )
            optimizerrr = self.optimizer(optim_list)
            aa = []
            for iter in tqdm(np.arange(self.iteration)):
                self.refresh_variable(self.dataset.Profiles.variableParam, 
                                      optim_list)
                conv_model = self.modeling(
                    self.dataset.Profiles.conv_profile_dict
                    )
                no_conv_model = self.modeling(
                    self.dataset.Profiles.no_conv_profile_dict
                    )
                model_tot = (self.add_convolution(conv_model.sum(axis=0)) 
                             + no_conv_model.sum(axis=0))

                loss, chi_mu = self.optimizing(self.data_cube[:, 0], 
                                                model_tot, 
                                                self.data_cube[:, 1])
                loss.backward()
                optimizerrr.step()
                optimizerrr.zero_grad()
                update, fitting_metrics = self.detect_update(fitting_metrics,
                                                        chi_mu, 
                                                        iter
                                                    ) 
                aa.append(loss.detach().cpu().numpy())
                if not update:
                    break 

            self.fitting_metrics_append(fitting_metrics[1], fitting_metrics[2]) 
            self.append_param_best_value()
            
            if isinstance(self.dataset.img_block_path, str):
                self.save_image()
        
        if isinstance(self.dataset.result_path, str):
            self.save_param()    

        if self.dataset.img_block_path is not None:
            print("The image block is saved in {} successfully!"
                  .format(self.dataset.img_block_path))  
        else:
            print("The image block is not saved becuase"
                  " img_block_path is not set.")     
                
        if self.dataset.result_path is not None:
            print("The parameter fitting result is saved in {} successfully!"
                  .format(self.dataset.result_path))  
        else:
            print("The parameter fitting result becuase"
                  " result_path is not set.")   
      


  



